typedef union {
    int b_value;
    int i_value;
    char *s_value;
    struct bytearray a_value;
  } YYSTYPE;
#define	L_TRUE	258
#define	L_FALSE	259
#define	L_INTEGER	260
#define	L_STRING	261
#define	L_BYTEARRAY	262
#define	L_DOUBLE1	263
#define	L_SOUNDCARD	264
#define	L_MIXER	265
#define	L_ELEMENT	266
#define	L_SWITCH	267
#define	L_RAWDATA	268
#define	L_CONTROL	269
#define	L_PCM	270
#define	L_RAWMIDI	271
#define	L_PLAYBACK	272
#define	L_CAPTURE	273
#define	L_INPUT	274
#define	L_OUTPUT	275
#define	L_SWITCH1	276
#define	L_SWITCH2	277
#define	L_SWITCH3	278
#define	L_VOLUME1	279
#define	L_3D_EFFECT1	280
#define	L_ACCU3	281
#define	L_MUX1	282
#define	L_MUX2	283
#define	L_TONE_CONTROL1	284
#define	L_IEC958OCS	285
#define	L_3D	286
#define	L_RESET	287
#define	L_USER	288
#define	L_VALID	289
#define	L_DATA	290
#define	L_PROTECT	291
#define	L_PRE2	292
#define	L_FSUNLOCK	293
#define	L_TYPE	294
#define	L_GSTATUS	295
#define	L_ENABLE	296
#define	L_DISABLE	297
#define	L_SW	298
#define	L_MONO_SW	299
#define	L_WIDE	300
#define	L_VOLUME	301
#define	L_CENTER	302
#define	L_SPACE	303
#define	L_DEPTH	304
#define	L_DELAY	305
#define	L_DEPTH_REAR	306
#define	L_FEEDBACK	307
#define	L_BASS	308
#define	L_TREBLE	309


extern YYSTYPE yylval;
