/*
 *  version.h
 */

#define SND_UTIL_MAJOR		0
#define SND_UTIL_MINOR		5
#define SND_UTIL_SUBMINOR	9b
#define SND_UTIL_VERSION		((SND_UTIL_MAJOR<<16)|\
				 (SND_UTIL_MINOR<<8)|\
				  SND_UTIL_SUBMINOR)
#define SND_UTIL_VERSION_STR	"0.5.9b"

